﻿// Arduino Sketch zum Betrieb von bis zu 7 induktiven Näherungssensoren
// Beschreibung siehe www.coole-basteleien.de/naeherungssensor.html
// Autor: A. Meßmer
//
// macros for pin settings
#define SET_AS_OUTPUT(DirectionReg, BitPos) (DirectionReg |= (1 << BitPos))
#define SET_AS_INPUT(DirectionReg, BitPos) (DirectionReg &= ~(1 << BitPos))
#define PIN_IS_HIGH(PinReg, BitPos) ((PinReg & (1 <<  BitPos)) != 0)
#define SET_HIGH(PortReg, BitPos) (PortReg |= (1 <<  BitPos))
#define SET_LOW(PortReg, BitPos)  (PortReg &= ~(1 <<  BitPos))
#define TOGGLE(PortReg, BitPos)  (PortReg ^= (1 <<  BitPos))

#define TEN_TIMES_NOP asm("nop\n nop\n nop\n nop\n nop\n nop\n nop\n nop\n nop\n nop\n") // 0.625 uS waiting time when 16 MHz clock

#define INTERRUPT_INTERVAL 156;  // 156 yields 3.2 kHz interrupt frequency in CTC mode when timer prescale is 32
#define SAMPLE_INTERVAL 32       // nr. of timer interrupts before next sampling starts, 32 yields 100 Hz sampling rate 
#define NR_SENSORS 2             // 7 is maximum, only 7 of the 8 PORTC pins ar usable, Pin PC7 is only read
#define ZERO_MEASUREMENT 32

volatile uint8_t measurement[NR_SENSORS], measurement_filtered[NR_SENSORS];
volatile uint8_t comparator_count;              // Counter incremented by comparator interrupt, must be volatile

#define SAMPLING  // if defined measurement values for a certain sensor are put on serial interface
#ifdef SAMPLING      
volatile uint16_t k_global;
#define SAMPLES 512
volatile uint8_t m1[SAMPLES],  m2[SAMPLES];
#endif // SAMPLING

volatile uint8_t measuring_count = 0;
volatile boolean measurement_in_progress = false;

void StartTimer(){
  /***************** setup timer ****************/
  ASSR &= ~(1<<AS2); // ensure system clock, only necessary for timer 2 
  TCCR2A = TCCR2B = TIMSK2 = 0x00;  // disables also timer compare interrupt
  TCCR2A |= (1 << WGM21);  // turn on CTC mode
  TCCR2B &= ~(1 << CS22);
  TCCR2B |= (1 << CS21) | (1 << CS20);   // Set CS20, CS21 bit for /32 prescale i.e. 2 us / count
  OCR2A = INTERRUPT_INTERVAL;  // * 2us for 32 prescale
  
  TCNT2  = 0;              //initialize timer counter value to 0
  TIMSK2 |= (1 << OCIE2A); // enable timer compare interrupt  
 
}

//************************************************************************
//********************************* Setup: *******************************
//************************************************************************
void setup() {
  uint8_t i;
  uint16_t k;

  MCUCR |= (1<<PUD);
  
  /***************** setup comparator ****************/
  DIDR1 = (1<<AIN1D) | (1<<AIN0D);  // Disable the digital input buffers.
  ACSR = 0;                         // comparator enabled, i.e. disable bit cleared, no bandgap, no input capture
  ACSR |= (1<<ACIS1);               // comparator interrupt on falling edge, since sensor is on negative comparator input this means rising for signal
  ACSR &= ~(1<<ACIE);               // comparator interrupt disabled
  ACSR |= (1<<ACI);                 // clear possible comparator flag
  /***************** setup the multiplexer ****************/
  ADCSRA &= ~(1<<ADEN); // ADC ausgeschalten (ADEN in ADCSRA Null)
  ADCSRB |= (1<<ACME); // Analog Comparator Multiplexer Enable Bit (ACME in ADCSRB Register) setzen
  ADMUX = 0x00; // MUX initial
  /***************** setup the analog pins ****************/
  DDRC = 0x00; // sets PC0...7 (ADC0...7), i.e. pin 14... as input (i.e. tri-state)
  measuring_count = 0;
  measurement_in_progress = false;
    
  SET_AS_OUTPUT(DDRB, DDB5); // LED Pin 13 for display purposes
  
  StartTimer();               // timer setup
  sei();                      // Enable interrupts

#ifdef SAMPLING      
  Serial.begin(57600); // start serial for output
#endif // SAMPLING

}

//************************************************************************
//********************************* loop: *******************************
//************************************************************************
void loop() {

/* here you can put your control program using measurement[] */


#ifdef SAMPLING      
  uint16_t k;
  k_global = 0;

  delay(100);
// wait for trigger condition
  while (measurement[0] > ZERO_MEASUREMENT-4) k_global = 0;

  while (k_global < (SAMPLES));   // wait until buffer is filled
  TCCR2A = TCCR2B = TIMSK2 = 0x00;  // disables also timer compare interrupt
  
  for (k = 0; k < SAMPLES; k++) {
    Serial.print((float)m1[k],0);
    Serial.print("  ");
    Serial.print((float)m2[k],0);
    Serial.print("\n");
    Serial.flush();
  }
  while(1); // stop here, output is found on Serial Monitor
#endif // SAMPLING
}  // main loop

  
ISR(TIMER2_COMPA_vect, ISR_NOBLOCK){
uint8_t bitmask, channel, m_smooth;

  /* measurements */
  channel = ((measuring_count++) % SAMPLE_INTERVAL);

  if (measurement_in_progress) { // terminate and save the measurement started for the previous channel
    channel--; // the previous channel
    ACSR &= ~(1<<ACIE);  // disable interrupt of analog comparator, result is in "comparator_count"
    measurement_in_progress = false;

    measurement[channel] = comparator_count;
    // smooth measurement by 1st order filter
    m_smooth = measurement_filtered[channel];
    m_smooth += min(comparator_count, ZERO_MEASUREMENT); // addd new result, counts above ZERO_MEASUREMENT (due to HF disturbances) are cut off
    m_smooth >>= 1;
      
    measurement_filtered[channel] = m_smooth;

    channel++; // increment again to the current channel
  } // saving of previous measurement completed
  
  if (channel < NR_SENSORS) { // if still a channel to serve
    measurement_in_progress = true;
    ADMUX = channel & ((1<<MUX2)|(1<<MUX1)|(1<<MUX0)); // MUX2 bis MUX0 Bits im ADMUX Register      
    bitmask = (1<<channel); // for selecting pin PC0...7
    /******** start measuring sequence ********/
    cli();
    
    /* pull down pulse */
    PORTC &= ~bitmask; // sets PC0...7, i.e. pin 14... to LOW 
    DDRC |= bitmask;  // sets PC0...7 (ADC0...7), i.e. pin 14... as output (i.e. pull down)
    TEN_TIMES_NOP; TEN_TIMES_NOP;  // 1.25 uS pulse duration
    DDRC &= ~bitmask;  // sets  PC0...7 (ADC0...7), i.e. pin 14... as input (i.e. tri-state)
    
    TEN_TIMES_NOP;                // 0.625 uS pause
    
    /* pull up pulse */
    PORTC |= bitmask; // sets PC0...7, i.e. pin 14... to HIGH 
    DDRC |= bitmask;  // sets PC0...7 (ADC0...7), i.e. pin 14... as output (i.e. pull up)
    TEN_TIMES_NOP; TEN_TIMES_NOP;  // 1.25 uS counter pulse duration
    DDRC &= ~bitmask;  // sets  PC0...7 (ADC0...7), i.e. pin 14... as input (i.e. tri-state)

    /* start counting */
    comparator_count = 0;
    sei();
    ACSR |= (1<<ACIE) | (1<<ACI);  // enable interrupt for analog comparator and delete possible comparator flag
  }  // for current channel

  
#ifdef SAMPLING      
  if ((channel) == NR_SENSORS) {
    m1[k_global] = measurement[0];  // be shure that the chosen measurement index is below NR_SENSORS
    m2[k_global] = measurement[1];
    k_global++;
    if (k_global > SAMPLES) k_global = 0;
  }
#endif // SAMPLING 
 
}

ISR(ANALOG_COMP_vect) {

  comparator_count++;
}
   


